@@ -12,6 +12,7 @@ from .base import (
     NewConnectionRequired,
 )
 from .http import AsyncBaseHTTPConnection
+from .http11 import AsyncHTTP11Connection
 
 logger = get_logger(__name__)
 
@@ -71,17 +72,16 @@ class AsyncHTTPConnection(AsyncHTTPTransport):
             self._request_lock = self.backend.create_lock()
         return self._request_lock
 
-    async def arequest(
+    async def handle_async_request(
         self,
         method: bytes,
         url: URL,
-        headers: Headers = None,
-        stream: AsyncByteStream = None,
-        ext: dict = None,
+        headers: Headers,
+        stream: AsyncByteStream,
+        extensions: dict,
     ) -> Tuple[int, Headers, AsyncByteStream, dict]:
         assert url_to_origin(url) == self.origin
-        ext = {} if ext is None else ext
-        timeout = cast(TimeoutDict, ext.get("timeout", {}))
+        timeout = cast(TimeoutDict, extensions.get("timeout", {}))
 
         async with self.request_lock:
             if self.state == ConnectionState.PENDING:
@@ -100,9 +100,14 @@ class AsyncHTTPConnection(AsyncHTTPTransport):
 
         assert self.connection is not None
         logger.trace(
-            "connection.arequest method=%r url=%r headers=%r", method, url, headers
+            "connection.handle_async_request method=%r url=%r headers=%r",
+            method,
+            url,
+            headers,
+        )
+        return await self.connection.handle_async_request(
+            method, url, headers, stream, extensions
         )
-        return await self.connection.arequest(method, url, headers, stream, ext)
 
     async def _open_socket(self, timeout: TimeoutDict = None) -> AsyncSocketStream:
         scheme, hostname, port = self.origin
@@ -150,8 +155,6 @@ class AsyncHTTPConnection(AsyncHTTPTransport):
                 socket=socket, backend=self.backend, ssl_context=self.ssl_context
             )
         else:
-            from .http11 import AsyncHTTP11Connection
-
             self.is_http11 = True
             self.connection = AsyncHTTP11Connection(
                 socket=socket, ssl_context=self.ssl_context
