@@ -1,8 +1,10 @@
+from http import HTTPStatus
 from ssl import SSLContext
-from typing import Tuple
+from typing import Tuple, cast
 
+from .._bytestreams import ByteStream
 from .._exceptions import ProxyError
-from .._types import URL, Headers, Origin, TimeoutDict
+from .._types import URL, Headers, TimeoutDict
 from .._utils import get_logger, url_to_origin
 from .base import AsyncByteStream
 from .connection import AsyncHTTPConnection
@@ -11,11 +13,19 @@ from .connection_pool import AsyncConnectionPool, ResponseByteStream
 logger = get_logger(__name__)
 
 
+def get_reason_phrase(status_code: int) -> str:
+    try:
+        return HTTPStatus(status_code).phrase
+    except ValueError:
+        return ""
+
+
 def merge_headers(
     default_headers: Headers = None, override_headers: Headers = None
 ) -> Headers:
     """
-    Append default_headers and override_headers, de-duplicating if a key existing in both cases.
+    Append default_headers and override_headers, de-duplicating if a key existing in
+    both cases.
     """
     default_headers = [] if default_headers is None else default_headers
     override_headers = [] if override_headers is None else override_headers
@@ -32,21 +42,23 @@ class AsyncHTTPProxy(AsyncConnectionPool):
     """
     A connection pool for making HTTP requests via an HTTP proxy.
 
-    **Parameters:**
-
-    * **proxy_url** - `Tuple[bytes, bytes, Optional[int], bytes]` - The URL of
-    the proxy service as a 4-tuple of (scheme, host, port, path).
-    * **proxy_headers** - `Optional[List[Tuple[bytes, bytes]]]` - A list of
-    proxy headers to include.
-    * **proxy_mode** - `str` - A proxy mode to operate in. May be "DEFAULT",
-    "FORWARD_ONLY", or "TUNNEL_ONLY".
-    * **ssl_context** - `Optional[SSLContext]` - An SSL context to use for
-    verifying connections.
-    * **max_connections** - `Optional[int]` - The maximum number of concurrent
-    connections to allow.
-    * **max_keepalive_connections** - `Optional[int]` - The maximum number of
-    connections to allow before closing keep-alive connections.
-    * **http2** - `bool` - Enable HTTP/2 support.
+    Parameters
+    ----------
+    proxy_url:
+        The URL of the proxy service as a 4-tuple of (scheme, host, port, path).
+    proxy_headers:
+        A list of proxy headers to include.
+    proxy_mode:
+        A proxy mode to operate in. May be "DEFAULT", "FORWARD_ONLY", or "TUNNEL_ONLY".
+    ssl_context:
+        An SSL context to use for verifying connections.
+    max_connections:
+        The maximum number of concurrent connections to allow.
+    max_keepalive_connections:
+        The maximum number of connections to allow before closing keep-alive
+        connections.
+    http2:
+        Enable HTTP/2 support.
     """
 
     def __init__(
@@ -59,6 +71,7 @@ class AsyncHTTPProxy(AsyncConnectionPool):
         max_keepalive_connections: int = None,
         keepalive_expiry: float = None,
         http2: bool = False,
+        backend: str = "auto",
         # Deprecated argument style:
         max_keepalive: int = None,
     ):
@@ -73,17 +86,18 @@ class AsyncHTTPProxy(AsyncConnectionPool):
             max_keepalive_connections=max_keepalive_connections,
             keepalive_expiry=keepalive_expiry,
             http2=http2,
+            backend=backend,
             max_keepalive=max_keepalive,
         )
 
-    async def request(
+    async def handle_async_request(
         self,
         method: bytes,
         url: URL,
-        headers: Headers = None,
-        stream: AsyncByteStream = None,
-        timeout: TimeoutDict = None,
-    ) -> Tuple[bytes, int, bytes, Headers, AsyncByteStream]:
+        headers: Headers,
+        stream: AsyncByteStream,
+        extensions: dict,
+    ) -> Tuple[int, Headers, AsyncByteStream, dict]:
         if self._keepalive_expiry is not None:
             await self._keepalive_sweep()
 
@@ -99,7 +113,7 @@ class AsyncHTTPProxy(AsyncConnectionPool):
                 url,
             )
             return await self._forward_request(
-                method, url, headers=headers, stream=stream, timeout=timeout
+                method, url, headers=headers, stream=stream, extensions=extensions
             )
         else:
             # By default HTTPS should be tunnelled.
@@ -111,29 +125,30 @@ class AsyncHTTPProxy(AsyncConnectionPool):
                 url,
             )
             return await self._tunnel_request(
-                method, url, headers=headers, stream=stream, timeout=timeout
+                method, url, headers=headers, stream=stream, extensions=extensions
             )
 
     async def _forward_request(
         self,
         method: bytes,
         url: URL,
-        headers: Headers = None,
-        stream: AsyncByteStream = None,
-        timeout: TimeoutDict = None,
-    ) -> Tuple[bytes, int, bytes, Headers, AsyncByteStream]:
+        headers: Headers,
+        stream: AsyncByteStream,
+        extensions: dict,
+    ) -> Tuple[int, Headers, AsyncByteStream, dict]:
         """
         Forwarded proxy requests include the entire URL as the HTTP target,
         rather than just the path.
         """
+        timeout = cast(TimeoutDict, extensions.get("timeout", {}))
         origin = self.proxy_origin
         connection = await self._get_connection_from_pool(origin)
 
         if connection is None:
             connection = AsyncHTTPConnection(
-                origin=origin, http2=self._http2, ssl_context=self._ssl_context,
+                origin=origin, http2=self._http2, ssl_context=self._ssl_context
             )
-            await self._add_to_pool(connection)
+            await self._add_to_pool(connection, timeout)
 
         # Issue a forwarded proxy request...
 
@@ -150,38 +165,38 @@ class AsyncHTTPProxy(AsyncConnectionPool):
         headers = merge_headers(self.proxy_headers, headers)
 
         (
-            http_version,
             status_code,
-            reason_phrase,
             headers,
             stream,
-        ) = await connection.request(
-            method, url, headers=headers, stream=stream, timeout=timeout
+            extensions,
+        ) = await connection.handle_async_request(
+            method, url, headers=headers, stream=stream, extensions=extensions
         )
 
         wrapped_stream = ResponseByteStream(
             stream, connection=connection, callback=self._response_closed
         )
 
-        return http_version, status_code, reason_phrase, headers, wrapped_stream
+        return status_code, headers, wrapped_stream, extensions
 
     async def _tunnel_request(
         self,
         method: bytes,
         url: URL,
-        headers: Headers = None,
-        stream: AsyncByteStream = None,
-        timeout: TimeoutDict = None,
-    ) -> Tuple[bytes, int, bytes, Headers, AsyncByteStream]:
+        headers: Headers,
+        stream: AsyncByteStream,
+        extensions: dict,
+    ) -> Tuple[int, Headers, AsyncByteStream, dict]:
         """
         Tunnelled proxy requests require an initial CONNECT request to
         establish the connection, and then send regular requests.
         """
+        timeout = cast(TimeoutDict, extensions.get("timeout", {}))
         origin = url_to_origin(url)
         connection = await self._get_connection_from_pool(origin)
 
         if connection is None:
-            scheme, host, port, _ = url
+            scheme, host, port = origin
 
             # First, create a connection to the proxy server
             proxy_connection = AsyncHTTPConnection(
@@ -194,40 +209,47 @@ class AsyncHTTPProxy(AsyncConnectionPool):
 
             # CONNECT www.example.org:80 HTTP/1.1
             # [proxy-headers]
-            if port is None:
-                target = host
-            else:
-                target = b"%b:%d" % (host, port)
+            target = b"%b:%d" % (host, port)
             connect_url = self.proxy_origin + (target,)
             connect_headers = [(b"Host", target), (b"Accept", b"*/*")]
             connect_headers = merge_headers(connect_headers, self.proxy_headers)
-            (
-                _,
-                proxy_status_code,
-                proxy_reason_phrase,
-                _,
-                proxy_stream,
-            ) = await proxy_connection.request(
-                b"CONNECT", connect_url, headers=connect_headers, timeout=timeout
-            )
-            logger.trace(
-                "tunnel_response proxy_status_code=%r proxy_reason=%r ",
-                proxy_status_code,
-                proxy_reason_phrase,
-            )
-            # Read the response data without closing the socket
-            async for _ in proxy_stream:
-                pass
 
-            # See if the tunnel was successfully established.
-            if proxy_status_code < 200 or proxy_status_code > 299:
-                msg = "%d %s" % (proxy_status_code, proxy_reason_phrase.decode("ascii"))
-                raise ProxyError(msg)
+            try:
+                (
+                    proxy_status_code,
+                    _,
+                    proxy_stream,
+                    _,
+                ) = await proxy_connection.handle_async_request(
+                    b"CONNECT",
+                    connect_url,
+                    headers=connect_headers,
+                    stream=ByteStream(b""),
+                    extensions=extensions,
+                )
 
-            # Upgrade to TLS if required
-            # We assume the target speaks TLS on the specified port
-            if scheme == b"https":
-                await proxy_connection.start_tls(host, timeout)
+                proxy_reason = get_reason_phrase(proxy_status_code)
+                logger.trace(
+                    "tunnel_response proxy_status_code=%r proxy_reason=%r ",
+                    proxy_status_code,
+                    proxy_reason,
+                )
+                # Read the response data without closing the socket
+                async for _ in proxy_stream:
+                    pass
+
+                # See if the tunnel was successfully established.
+                if proxy_status_code < 200 or proxy_status_code > 299:
+                    msg = "%d %s" % (proxy_status_code, proxy_reason)
+                    raise ProxyError(msg)
+
+                # Upgrade to TLS if required
+                # We assume the target speaks TLS on the specified port
+                if scheme == b"https":
+                    await proxy_connection.start_tls(host, timeout)
+            except Exception as exc:
+                await proxy_connection.aclose()
+                raise ProxyError(exc)
 
             # The CONNECT request is successful, so we have now SWITCHED PROTOCOLS.
             # This means the proxy connection is now unusable, and we must create
@@ -239,22 +261,25 @@ class AsyncHTTPProxy(AsyncConnectionPool):
                 ssl_context=self._ssl_context,
                 socket=proxy_connection.socket,
             )
-            await self._add_to_pool(connection)
+            await self._add_to_pool(connection, timeout)
 
         # Once the connection has been established we can send requests on
         # it as normal.
         (
-            http_version,
             status_code,
-            reason_phrase,
             headers,
             stream,
-        ) = await connection.request(
-            method, url, headers=headers, stream=stream, timeout=timeout,
+            extensions,
+        ) = await connection.handle_async_request(
+            method,
+            url,
+            headers=headers,
+            stream=stream,
+            extensions=extensions,
         )
 
         wrapped_stream = ResponseByteStream(
             stream, connection=connection, callback=self._response_closed
         )
 
-        return http_version, status_code, reason_phrase, headers, wrapped_stream
+        return status_code, headers, wrapped_stream, extensions
